"use strict";
// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
const jupyter_dataserializers_1 = require("jupyter-dataserializers");
const base_2 = require("jupyter-datawidgets/lib/base");
const colormap_1 = require("./colormap");
const version_1 = require("./version");
const utils_1 = require("./utils");
const ndarray = require("ndarray");
/**
 * Create new ndarray, with default attributes taken from another
 *
 * @param {ndarray.NDArray} array
 * @returns {ndarray.NDArray}
 */
function arrayFrom(array, dtype, shape) {
    dtype = dtype || array.dtype;
    shape = shape || array.shape;
    if (dtype === 'buffer' || dtype === 'generic' || dtype === 'array') {
        throw new Error(`Cannot create ndarray of dtype "${dtype}".`);
    }
    return ndarray(new jupyter_dataserializers_1.typesToArray[dtype](shape.reduce((ac, v) => {
        ac *= v;
        return ac;
    }, 1)), shape, array.stride, array.offset);
}
exports.arrayFrom = arrayFrom;
/**
 * Whether two ndarrays differ in shape.
 */
function shapesDiffer(a, b) {
    if (a === null && b === null) {
        return false;
    }
    return a === null || b === null ||
        JSON.stringify(a) !== JSON.stringify(b);
}
/**
 * Scaled array model.
 *
 * This model provides a scaled version of an array, that is
 * automatically recomputed when either the array or the scale
 * changes.
 *
 * @export
 * @class ScaledArrayModel
 * @extends {DataModel}
 */
class ScaledArrayModel extends base_2.DataModel {
    defaults() {
        const ctor = this.constructor;
        return Object.assign({}, super.defaults(), {
            _model_name: ctor.model_name,
            _model_module: ctor.model_module,
            _model_module_version: ctor.model_module_version,
            _view_name: ctor.view_name,
            _view_module: ctor.view_module,
            _view_module_version: ctor.view_module_version,
            data: ndarray([]),
            scale: null,
            scaledData: null,
            output_dtype: 'inherit',
        });
    }
    /**
     * (Re-)compute the scaledData data.
     *
     * @returns {void}
     * @memberof ScaledArrayModel
     */
    computeScaledData(options) {
        options = typeof options === 'object'
            ? Object.assign({}, options, { setScaled: true }) : { setScaled: true };
        let array = jupyter_dataserializers_1.getArray(this.get('data'));
        let scale = this.get('scale');
        // Handle null case immediately:
        if (array === null || scale === null) {
            this.set('scaledData', null, options);
            return;
        }
        let resized = this.arrayMismatch();
        let scaledData = this.get('scaledData');
        if (resized) {
            // Allocate new array
            scaledData = arrayFrom(array, this.scaledDtype(), this.scaledShape());
        }
        else {
            // Reuse data, but wrap in new ndarray object to trigger change
            const version = scaledData._version + 1 || 0;
            scaledData = ndarray(scaledData.data, scaledData.shape, scaledData.stride, scaledData.offset);
            // Tag on a version# to differntiate it:
            scaledData._version = version;
        }
        let data = array.data;
        let target = scaledData.data;
        // Set values:
        if (colormap_1.isColorMapModel(scale)) {
            for (let i = 0; i < data.length; ++i) {
                const c = utils_1.parseCssColor(scale.obj(data[i]));
                target[i * 4 + 0] = c[0];
                target[i * 4 + 1] = c[1];
                target[i * 4 + 2] = c[2];
                target[i * 4 + 3] = c[3];
            }
        }
        else {
            for (let i = 0; i < data.length; ++i) {
                target[i] = scale.obj(data[i]);
            }
        }
        this.set('scaledData', scaledData, options);
    }
    /**
     * Initialize the model
     *
     * @param {Backbone.ObjectHash} attributes
     * @param {{model_id: string; comm?: any; widget_manager: any; }} options
     * @memberof ScaledArrayModel
     */
    initialize(attributes, options) {
        super.initialize(attributes, options);
        const scale = this.get('scale') || undefined;
        // Await scale object for init:
        this.initPromise = Promise.resolve(scale && scale.initPromise).then(() => {
            this.computeScaledData();
            this.setupListeners();
        });
    }
    /**
     * Sets up any relevant event listeners after the object has been initialized,
     * but before the initPromise is resolved.
     *
     * @memberof ScaledArrayModel
     */
    setupListeners() {
        // Listen to direct changes on our model:
        this.on('change:scale', this.onChange, this);
        // Listen to changes within array and scale models:
        jupyter_dataserializers_1.listenToUnion(this, 'data', this.onChange.bind(this), true);
        this.listenTo(this.get('scale'), 'change', this.onChange);
        // make sure to (un)hook listeners when child points to new object
        this.on('change:scale', (model, value, options) => {
            const prevModel = this.previous('scale');
            const currModel = value;
            if (prevModel) {
                this.stopListening(prevModel);
            }
            if (currModel) {
                this.listenTo(currModel, 'change', this.onChange.bind(this));
            }
        }, this);
    }
    getNDArray(key = 'scaledData') {
        if (key === 'scaledData') {
            if (this.get('scaledData') === null) {
                this.computeScaledData();
            }
            return this.get('scaledData');
        }
        else {
            return this.get(key);
        }
    }
    canWriteBack(key = 'scaledData') {
        if (key === 'data') {
            return true;
        }
        if (key !== 'scaledData') {
            return false;
        }
        const scale = this.get('scale');
        if (colormap_1.isColorMapModel(scale)) {
            return false;
        }
        return !!scale && typeof scale.obj.invert === 'function';
    }
    setNDArray(array, key = 'scaledData', options) {
        if (key === 'scaledData') {
            // Writing back, we need to feed the data through scale.invert()
            const current = jupyter_dataserializers_1.getArray(this.get('data'));
            const scale = this.get('scale');
            // Handle null case immediately:
            if (array === null || scale === null) {
                jupyter_dataserializers_1.setArray(this, 'data', null, options);
                return;
            }
            // Allocate new array
            const dtype = current ? current.dtype : array.dtype;
            // Special case colors, as we allow them to transform the shape
            const shape = array.shape;
            const newArray = arrayFrom(array, dtype, shape);
            let data = array.data;
            let target = newArray.data;
            // Set values:
            for (let i = 0; i < data.length; ++i) {
                target[i] = scale.obj.invert(data[i]);
            }
            jupyter_dataserializers_1.setArray(this, 'data', newArray, options);
        }
        else {
            jupyter_dataserializers_1.setArray(this, key, array, options);
        }
    }
    /**
     * Callback for when the source data changes.
     *
     * @param {WidgetModel} model
     * @memberof ScaledArrayModel
     */
    onChange(model, options) {
        if (!options || options.setScaled !== true) {
            this.computeScaledData(options);
        }
    }
    /**
     * Whether scaledData has the incorrect shape or type.
     *
     * @protected
     * @returns {boolean}
     * @memberof ScaledArrayModel
     */
    arrayMismatch() {
        const current = this.get('scaledData');
        if (current && current.dtype !== this.scaledDtype()) {
            return true;
        }
        return shapesDiffer(current && current.shape, this.scaledShape());
    }
    /**
     * Get what the dtype of the scaled data *should* be
     */
    scaledDtype() {
        let output_dtype = this.get('output_dtype');
        if (output_dtype !== 'inherit') {
            return output_dtype;
        }
        let array = jupyter_dataserializers_1.getArray(this.get('data'));
        if (array === null) {
            return null;
        }
        return array.dtype;
    }
    /**
     * Get what the shape of the scaled data *should* be
     */
    scaledShape() {
        const scale = this.get('scale');
        const array = jupyter_dataserializers_1.getArray(this.get('data'));
        // Special case colors, as we allow them to transform the shape
        if (colormap_1.isColorMapModel(scale)) {
            return array && array.shape.concat(4);
        }
        return array && array.shape;
    }
}
ScaledArrayModel.serializers = Object.assign({}, base_2.DataModel.serializers, { data: jupyter_dataserializers_1.data_union_serialization, scale: { deserialize: base_1.unpack_models }, scaledData: { serialize: utils_1.undefSerializer } });
ScaledArrayModel.model_name = 'ScaledArrayModel';
ScaledArrayModel.model_module = version_1.moduleName;
ScaledArrayModel.model_module_version = version_1.version;
exports.ScaledArrayModel = ScaledArrayModel;
//# sourceMappingURL=datawidgets.js.map