"use strict";
// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
const jupyter_dataserializers_1 = require("jupyter-dataserializers");
const version_1 = require("./version");
const utils_1 = require("./utils");
/**
 * Scaled value model.
 *
 * This model provides a scaled value, that is automatically recomputed when
 * either the input value or the scale changes.
 */
class ScaledValueModel extends base_1.WidgetModel {
    defaults() {
        const ctor = this.constructor;
        return Object.assign({}, super.defaults(), {
            _model_name: ctor.model_name,
            _model_module: ctor.model_module,
            _model_module_version: ctor.model_module_version,
            _view_name: ctor.view_name,
            _view_module: ctor.view_module,
            _view_module_version: ctor.view_module_version,
            input: null,
            scale: null,
            output: null,
        });
    }
    /**
     * (Re-)compute the output.
     *
     * @returns {void}
     * @memberof ScaledArrayModel
     */
    computeScaledValue(options) {
        options = typeof options === 'object'
            ? Object.assign({}, options, { setOutputOf: this }) : { setOutput: true };
        let scale = this.get('scale');
        let input = this.get('input');
        // If input is another ScaledValueModel, use its output as our input:
        if (input instanceof ScaledValueModel) {
            input = input.get('output');
        }
        // Handle null case immediately:
        if (input === null || scale === null) {
            this.set('output', null, options);
            return;
        }
        this.set('output', scale.obj(input), options);
    }
    /**
     * Initialize the model
     *
     * @param {Backbone.ObjectHash} attributes
     * @param {{model_id: string; comm?: any; widget_manager: any; }} options
     * @memberof ScaledArrayModel
     */
    initialize(attributes, options) {
        super.initialize(attributes, options);
        const scale = this.get('scale') || undefined;
        // Await scale object for init:
        this.initPromise = Promise.resolve(scale && scale.initPromise).then(() => {
            this.computeScaledValue();
            this.setupListeners();
        });
    }
    /**
     * Sets up any relevant event listeners after the object has been initialized,
     * but before the initPromise is resolved.
     *
     * @memberof ScaledArrayModel
     */
    setupListeners() {
        // Listen to changes on scale model:
        this.listenTo(this.get('scale'), 'change', this.onChange);
        // make sure to (un)hook listeners when child points to new object
        this.on('change:scale', (model, value, options) => {
            const prevModel = this.previous('scale');
            const currModel = value;
            if (prevModel) {
                this.stopListening(prevModel);
            }
            if (currModel) {
                this.listenTo(currModel, 'change', this.onChange.bind(this));
            }
            this.onChange(this);
        }, this);
        // Listen to changes on input union:
        jupyter_dataserializers_1.listenToUnion(this, 'input', this.onChange.bind(this), true);
    }
    /**
     * Callback for when the source input changes.
     *
     * @param {WidgetModel} model
     * @memberof ScaledArrayModel
     */
    onChange(model, options) {
        if (!options || options.setOutputOf !== this) {
            this.computeScaledValue(options);
        }
    }
}
ScaledValueModel.serializers = {
    input: { deserialize: base_1.unpack_models },
    scale: { deserialize: base_1.unpack_models },
    output: { serialize: utils_1.undefSerializer },
};
ScaledValueModel.model_name = 'ScaledValueModel';
ScaledValueModel.model_module = version_1.moduleName;
ScaledValueModel.model_module_version = version_1.version;
ScaledValueModel.view_name = null;
ScaledValueModel.view_module = null;
ScaledValueModel.view_module_version = version_1.version;
exports.ScaledValueModel = ScaledValueModel;
//# sourceMappingURL=value.js.map